package com.github.pfichtner.durationformatter;

import static com.google.common.base.Preconditions.checkState;
import static java.util.concurrent.TimeUnit.DAYS;
import static java.util.concurrent.TimeUnit.HOURS;
import static java.util.concurrent.TimeUnit.MICROSECONDS;
import static java.util.concurrent.TimeUnit.MILLISECONDS;
import static java.util.concurrent.TimeUnit.MINUTES;
import static java.util.concurrent.TimeUnit.NANOSECONDS;
import static java.util.concurrent.TimeUnit.SECONDS;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.TimeUnit;

import javax.annotation.concurrent.ThreadSafe;

import com.google.common.base.Joiner;
import com.google.common.base.Throwables;

/**
 * A Formatter for durations. This class is threadsafe. Instances can be created
 * using the {@link Builder} class.
 * 
 * @author Peter Fichtner
 */
@ThreadSafe
public class DurationFormatter {

	/**
	 * Helper class to create {@link DurationFormatter}s. This class is
	 * threadsafe so each method call will return a new instance!
	 * 
	 * @author Peter Fichtner
	 */
	@ThreadSafe
	public static class Builder implements Cloneable {

		private static final Builder BASE = new Builder().minimum(SECONDS)
				.maximum(HOURS).format(TimeUnit.MICROSECONDS, "%03d")
				.format(TimeUnit.MICROSECONDS, "%03d")
				.format(TimeUnit.NANOSECONDS, "%03d");

		/**
		 * Default instance that formats seconds to hours using digits (e.g.
		 * <code>01:12:33</code>)
		 */
		public static final Builder DIGITS = BASE;

		/**
		 * Default instance that formats seconds to hours using digits (e.g.
		 * <code>01:12:33</code>)
		 */
		public static final Builder SYMBOLS = BASE.separator(" ").format("%d")
				.symbol(NANOSECONDS, "ns").symbol(MICROSECONDS, "µs")
				.symbol(MILLISECONDS, "ms").symbol(SECONDS, "s")
				.symbol(MINUTES, "min").symbol(HOURS, "h").symbol(DAYS, "d");

		private String separator = ":";
		private TimeUnit minimum = MILLISECONDS;
		private TimeUnit maximum = HOURS;
		private boolean round = true;
		private boolean stripLeadingZeros;
		private HashMap<TimeUnit, String> formats = new HashMap<TimeUnit, String>();
		private HashMap<TimeUnit, String> symbols = new HashMap<TimeUnit, String>();

		public DurationFormatter build() {
			return new DurationFormatter(this);
		}

		public Builder maximum(TimeUnit maximum) {
			Builder clone = clone();
			clone.maximum = maximum;
			return clone;
		}

		public Builder minimum(TimeUnit minimum) {
			Builder clone = clone();
			clone.minimum = minimum;
			return clone;
		}

		public Builder round(boolean round) {
			Builder clone = clone();
			clone.round = round;
			return clone;
		}

		public Builder separator(String separator) {
			Builder clone = clone();
			clone.separator = separator;
			return clone;
		}

		public Builder format(String format) {
			Builder clone = clone();
			for (TimeUnit timeUnit : timeUnits) {
				clone = clone.format(timeUnit, format);
			}
			return clone;
		}

		public Builder format(TimeUnit timeUnit, String format) {
			Builder clone = clone();
			clone.formats.put(timeUnit, format);
			return clone;
		}

		public Builder symbol(TimeUnit timeUnit, String symbol) {
			Builder clone = clone();
			clone.symbols.put(timeUnit, symbol);
			return clone;
		}

		public Builder stripLeadingZeros(boolean stripLeadingZeros) {
			Builder clone = clone();
			clone.stripLeadingZeros = stripLeadingZeros;
			return clone;
		}

		@Override
		protected Builder clone() {
			try {
				Builder clone = (Builder) super.clone();
				clone.formats = new HashMap<TimeUnit, String>(this.formats);
				clone.symbols = new HashMap<TimeUnit, String>(this.symbols);
				return clone;
			} catch (CloneNotSupportedException e) {
				throw Throwables.propagate(e);
			}
		}

	}

	private static final List<TimeUnit> timeUnits = Collections
			.unmodifiableList(new ArrayList<TimeUnit>(
					orderingNaturalReverse(Arrays.asList(TimeUnit.values()))));
	private static final TimeUnit highestPrecision = getLast(timeUnits);

	public static final DurationFormatter DIGITS = Builder.DIGITS.build();;

	public static final DurationFormatter SYMBOLS = Builder.SYMBOLS.build();

	private static final Integer ZERO = Integer.valueOf(0);

	private static final String DEFAULT_FORMAT = "%02d";

	private static String floor(long d, long n) {
		return Integer.toString((int) Math.floor((double) d / n));
	}

	private final boolean round;
	private final List<TimeUnit> usedTimeUnits;
	private final String separator;
	private int idxMin;
	private TimeUnit timeUnitMin;
	private Map<TimeUnit, String> formats;
	private Map<TimeUnit, String> symbols;
	private boolean stripLeadingZeros;

	public DurationFormatter(Builder builder) {
		this.round = builder.round;
		checkState(builder.minimum.compareTo(builder.maximum) <= 0);
		this.idxMin = indexOf(timeUnits, builder.minimum);
		int idxMax = indexOf(timeUnits, builder.maximum);
		checkState(this.idxMin > idxMax, "min must not be greater than max");
		this.timeUnitMin = timeUnits.get(this.idxMin);
		this.usedTimeUnits = timeUnits.subList(idxMax, this.idxMin + 1);
		this.separator = builder.separator;
		this.stripLeadingZeros = builder.stripLeadingZeros;
		this.formats = Collections
				.unmodifiableMap(new HashMap<TimeUnit, String>(builder.formats));
		this.symbols = Collections
				.unmodifiableMap(new HashMap<TimeUnit, String>(builder.symbols));
	}

	/**
	 * Format the passed milliseconds to the format specified.
	 * 
	 * @param value
	 *            the millis to format
	 * @return String containing the duration
	 * @see #format(long, TimeUnit)
	 */
	public String formatMillis(long value) {
		return format(value, MILLISECONDS);
	}

	/**
	 * Format the passed duration to the format specified.
	 * 
	 * @param value
	 *            the duration to format
	 * @return String containing the duration
	 */
	public String format(long value, TimeUnit timeUnit) {
		long nanos = NANOSECONDS.convert(value, timeUnit);
		List<String> values = getValues(this.round
				&& !highestPrecision.equals(this.timeUnitMin) ? calculateRounded(nanos)
				: nanos);
		return Joiner.on(this.separator).join(values);
	}

	private long calculateRounded(long value) {
		TimeUnit smaller = timeUnits.get(this.idxMin + 1);
		long add = smaller.convert(1, this.timeUnitMin) / 2;
		return value + NANOSECONDS.convert(add, smaller);
	}

	private List<String> getValues(long delta) {
		List<String> strings = new ArrayList<String>();
		long actual = delta;

		TimeUnit last = getLast(this.usedTimeUnits);
		boolean added = false;
		for (TimeUnit timeUnit : this.usedTimeUnits) {
			long longVal = timeUnit.toNanos(1);
			String format = this.formats.get(timeUnit);
			boolean fits = actual >= longVal;
			if (added || fits || !this.stripLeadingZeros
					|| timeUnit.equals(last)) {
				String value = String.format(format == null ? DEFAULT_FORMAT
						: format,
						fits ? Integer.valueOf(floor(actual, longVal)) : ZERO);
				String symbol = this.symbols.get(timeUnit);
				strings.add(symbol == null ? value : value + symbol);
				added = true;
			}
			actual %= longVal;
		}
		return strings;
	}

	// -------------------------------------------------------------------------
	// - methods primarily found in google guava but redefined to minimize jar -
	// - size -
	// -------------------------------------------------------------------------

	private static <T> T getLast(List<T> ts) {
		int size = ts.size();
		return size == 0 ? null : ts.get(size - 1);
	}

	private static <T> int indexOf(List<T> ts, Object search) {
		int i = 0;
		for (T t : ts) {
			if (search.equals(t)) {
				return i;
			}
			i++;
		}
		return -1;
	}

	private static <T> List<T> orderingNaturalReverse(List<T> ts) {
		List<T> result = new ArrayList<T>(ts);
		Collections.sort(result, Collections.<T> reverseOrder());
		return result;
	}

}
